"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RuleEngine = void 0;
const dependency_graph_1 = require("../dependency-graph/dependency-graph");
const builder_1 = require("./builder/builder");
const validation_type_1 = require("../validators/validation-type");
class RuleEngine {
    constructor() {
        this.propertyMap = {};
        this.dependencyGraph = new dependency_graph_1.DependencyGraph();
        this.validations = new WeakMap();
        this.dataLinks = new Map();
        this.snapshots = new Map();
        this.propertiesThatRequireAnEagerUpdate = [];
    }
    get properties() {
        return Object.values(this.propertyMap);
    }
    builder(options) {
        return new builder_1.Builder(options, this, this.dependencyGraph, this.propertyMap);
    }
    takeSnapShot(key = 'default') {
        const snap = {
            key,
            data: this.properties.map(p => p.exportData())
        };
        this.snapshots.set(key, snap);
        return snap;
    }
    restoreSnapShot(key = 'default') {
        const snap = this.snapshots.get(key);
        if (snap) {
            snap.data.forEach((d, i) => this.properties[i].importData(d));
        }
    }
    linkPropertyData(propertyA, propertyB) {
        const dataLinkKey = this.createDataLinkKey(propertyA, propertyB);
        if (this.dataLinks.has(dataLinkKey)) {
            return;
        }
        const listenerA = {
            updated: () => this.transferData(propertyA, propertyB)
        };
        const listenerB = {
            updated: () => this.transferData(propertyB, propertyA)
        };
        this.transferData(propertyA, propertyB);
        propertyA.registerValueChangedListener(listenerA);
        propertyB.registerValueChangedListener(listenerB);
        this.dataLinks.set(dataLinkKey, [listenerA, listenerB]);
    }
    unlinkPropertyData(propertyA, propertyB) {
        const dataLinkKey = this.createDataLinkKey(propertyA, propertyB);
        const listeners = this.dataLinks.get(dataLinkKey);
        if (listeners) {
            const [listenerA, listenerB] = listeners;
            propertyA.deregisterValueChangedListener(listenerA);
            propertyB.deregisterValueChangedListener(listenerB);
            this.dataLinks.delete(dataLinkKey);
        }
    }
    createDataLinkKey(propertyA, propertyB) {
        return propertyA.id < propertyB.id
            ? `${propertyA.id}<->${propertyB.id}`
            : `${propertyB.id}<->${propertyA.id}`;
    }
    transferData(fromProperty, toProperty) {
        toProperty.importData(fromProperty.exportData());
    }
    needsAnUpdate(mightHaveChanged) {
        this.dependencyGraph.traverseDepthFirst(mightHaveChanged.id, property => {
            property.needsAnUpdate(false);
            if (property.internallyRequiresEagerUpdate()
                && this.propertiesThatRequireAnEagerUpdate.every(p => p.id !== property.id)) {
                this.propertiesThatRequireAnEagerUpdate.push(property);
            }
        }, dependency => !!dependency.options.value);
        setTimeout(() => {
            const eagerProps = this.propertiesThatRequireAnEagerUpdate;
            this.propertiesThatRequireAnEagerUpdate = [];
            eagerProps.forEach((prop) => void this.updateValue(prop));
        }, 0);
    }
    updateValue(property) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const asyncDeps = this.dependencyGraph.getAsyncDependencies(property.id);
                if (asyncDeps === null || asyncDeps === void 0 ? void 0 : asyncDeps.length) {
                    yield Promise.all(asyncDeps.map(asyncDep => this.updateValue(asyncDep)));
                }
                yield property.internallyUpdate();
                this.hasBeenUpdated(property);
            }
            catch (err) {
                property.errorWhileUpdating(err);
            }
        });
    }
    hasBeenUpdated(property) {
        property.hasBeenUpdated();
        this.dependencyGraph.traverseDepthFirst(property.id, (prop, dep) => prop.dependencyHasBeenUpdated(dep), dependency => !dependency.options.value, true);
    }
    invalidateValidationResults(validators) {
        validators.forEach(v => {
            const vprocess = this.getValidationProcess(v);
            vprocess.isLastResultUpToDate = false;
        });
    }
    validate(validators) {
        return validators.map(validator => {
            const vprocess = this.getValidationProcess(validator);
            if (vprocess.isLastResultUpToDate) {
                return Promise.resolve(vprocess.lastValidationResult);
            }
            else if (vprocess.currentValidation) {
                return vprocess.currentValidation;
            }
            const validation = validator.validate(...validator.getValidatedProperties());
            if (validation instanceof Promise) {
                vprocess.currentValidation = validation.then(res => {
                    vprocess.currentValidation = undefined;
                    vprocess.lastValidationResult = res;
                    vprocess.isLastResultUpToDate = true;
                    return res;
                }, err => {
                    vprocess.currentValidation = undefined;
                    return [{
                            text: JSON.stringify(err),
                            type: validation_type_1.ValidationTypes.ErrorThrownDuringValidation
                        }];
                });
                return vprocess.currentValidation;
            }
            else {
                vprocess.lastValidationResult = validation;
                vprocess.isLastResultUpToDate = true;
                return Promise.resolve(validation);
            }
        });
    }
    getValidationProcess(validator) {
        const vprocess = this.validations.get(validator);
        if (vprocess) {
            return vprocess;
        }
        else {
            const newProcess = {
                isLastResultUpToDate: false
            };
            this.validations.set(validator, newProcess);
            return newProcess;
        }
    }
}
exports.RuleEngine = RuleEngine;
