"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractPropertyImpl = void 0;
const logger_1 = require("../util/logger/logger");
const assertion_error_1 = require("../util/assertions/assertion-error");
class AbstractPropertyImpl {
    constructor(updateHandler, backpressureConfig) {
        this.updateHandler = updateHandler;
        this.automaticallyUpdate = true;
        this.validationMessages = [];
        this.valueChangeListeners = [];
        this.backpressureConfig = backpressureConfig;
    }
    awaitAsyncUpdate() {
        if (this.needsToRecompute !== false && this.automaticallyUpdate) {
            return this.updateHandler.updateValue(this);
        }
        else {
            return this.updatedPromise();
        }
    }
    updatedPromise() {
        if (this.currentRecomputing) {
            return new Promise((resolve, reject) => {
                const ul = { resolve, reject };
                if (!this.updatedListeners) {
                    this.updatedListeners = [ul];
                }
                else {
                    this.updatedListeners.push(ul);
                }
            });
        }
        else {
            return Promise.resolve();
        }
    }
    nextRecomputingCount() {
        var _a;
        this.recomputingCount = (((_a = this.recomputingCount) !== null && _a !== void 0 ? _a : 0) % Number.MAX_SAFE_INTEGER) + 1;
        return this.recomputingCount;
    }
    checkAndTriggerUpdate() {
        if (this.needsToRecompute !== false && this.automaticallyUpdate) {
            void this.updateHandler.updateValue(this);
        }
    }
    getTriggerListener() {
        if (!this.triggerListener) {
            this.triggerListener = {
                onTriggered: () => {
                    if (this.needsToRecompute !== false) {
                        this.manuallyTriggered = true;
                        return this.updateHandler.updateValue(this);
                    }
                    else {
                        return Promise.resolve();
                    }
                }
            };
        }
        return this.triggerListener;
    }
    internallyUpdate() {
        if (this.needsToRecompute === false || (!this.automaticallyUpdate && !this.manuallyTriggered)) {
            return this.updatedPromise();
        }
        this.needsToRecompute = false;
        delete this.manuallyTriggered;
        if (this.isAsynchronous()) {
            if (this.currentRecomputing) {
                return this.handleBackpressure();
            }
            else {
                return this.handleAsyncUpdate(this.nextRecomputingCount(), false);
            }
        }
        else {
            this.internallySyncUpdate();
            return Promise.resolve();
        }
    }
    handleBackpressure() {
        if (!this.backpressureConfig) {
            throw new assertion_error_1.AssertionError('AbstractPropertyImpl: backpressureConfig needs to be defined');
        }
        switch (this.backpressureConfig.type) {
            case 'switch':
                if (this.backpressureConfig.debounceTime) {
                    return this.handleAsyncUpdateWithDebounceTime(this.backpressureConfig.debounceTime);
                }
                else {
                    return this.handleAsyncUpdate(this.nextRecomputingCount(), true);
                }
            case 'skip':
            default:
                return this.updatedPromise();
        }
    }
    handleAsyncUpdateWithDebounceTime(debounceTime) {
        return __awaiter(this, void 0, void 0, function* () {
            const recomputingCount = this.nextRecomputingCount();
            yield Promise.race([
                this.currentRecomputing,
                new Promise(resolve => setTimeout(() => resolve(), debounceTime))
            ]);
            return this.handleAsyncUpdate(recomputingCount, true);
        });
    }
    handleAsyncUpdate(recomputingCount, wasAlreadyProcessing) {
        if (recomputingCount !== this.recomputingCount) {
            logger_1.Logger.trace(() => `Property.handleAsyncUpdate ${this.id}: ${recomputingCount} is skipped (${this.recomputingCount})`);
            return this.updatedPromise();
        }
        logger_1.Logger.trace(() => `Property.handleAsyncUpdate ${this.id}: ${recomputingCount}`);
        const update = this.internallyAsyncUpdate();
        this.currentRecomputing = update.asyncPromise.then((value) => {
            var _a;
            logger_1.Logger.trace(() => `Property.handleAsyncUpdate ${this.id}: ${recomputingCount} / ${this.recomputingCount}`);
            if (recomputingCount === this.recomputingCount) {
                logger_1.Logger.trace(() => `Property.handleAsyncUpdate ${this.id}: updated`);
                this.currentRecomputing = undefined;
                update.resolve(value);
                (_a = this.updatedListeners) === null || _a === void 0 ? void 0 : _a.forEach(ul => ul.resolve());
                delete this.updatedListeners;
            }
            else if (this.currentRecomputing) {
                return this.currentRecomputing;
            }
        });
        if (this.isProcessing() && !wasAlreadyProcessing) {
            this.tellValueChangeListeners(listener => {
                if (listener.startsAsyncUpdate) {
                    listener.startsAsyncUpdate();
                }
            });
        }
        return this.updatedPromise();
    }
    needsAnUpdate(notifyOthers) {
        logger_1.Logger.trace(() => `Property.needsAnUpdate ${this.id}`);
        this.needsToRecompute = true;
        this.needsToRevalidate = true;
        if (this.validators) {
            this.updateHandler.invalidateValidationResults(this.validators);
        }
        if (notifyOthers !== false) {
            this.updateHandler.needsAnUpdate(this);
        }
        this.tellValueChangeListeners(listener => {
            if (listener.needsAnUpdate) {
                listener.needsAnUpdate();
            }
        });
    }
    hasBeenUpdated() {
        this.tellValueChangeListeners(listener => listener.updated());
    }
    errorWhileUpdating(error) {
        var _a;
        (_a = this.updatedListeners) === null || _a === void 0 ? void 0 : _a.forEach(ul => ul.reject(error));
        delete this.updatedListeners;
        this.tellValueChangeListeners(listener => {
            if (listener.updateFailed) {
                listener.updateFailed(error);
            }
        });
    }
    dependencyHasBeenUpdated(dependency) {
        if (dependency.options.validation) {
            this.needsToRevalidate = true;
            if (this.validators) {
                this.updateHandler.invalidateValidationResults(this.validators);
            }
        }
        this.tellValueChangeListeners(listener => {
            if (listener.dependencyHasBeenUpdated) {
                listener.dependencyHasBeenUpdated(dependency);
            }
        });
    }
    internallyRequiresEagerUpdate() {
        return this.valueChangeListeners.length > 0;
    }
    addValidator(validator) {
        if (!this.validators) {
            this.validators = [];
        }
        this.validators.push(validator);
    }
    validate() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.needsToRevalidate !== false) {
                this.needsToRevalidate = false;
                yield this.awaitAsyncUpdate();
                this.validationMessages = this.getSpecialisedValidationResult();
                if (this.validators) {
                    const results = yield Promise.all(this.updateHandler.validate(this.validators));
                    results.forEach(result => {
                        if (result instanceof Array) {
                            this.validationMessages.push(...result);
                        }
                        else if (!!result && result[this.id]) {
                            this.validationMessages.push(...result[this.id]);
                        }
                    });
                }
                this.tellValueChangeListeners(listener => {
                    if (listener.validated) {
                        listener.validated();
                    }
                });
            }
            return this.getValidationMessages();
        });
    }
    isValid() {
        return this.validationMessages.every(msg => msg.type.isValid);
    }
    getValidationMessages() {
        return this.validationMessages;
    }
    registerValueChangedListener(changed) {
        this.valueChangeListeners.push(changed);
        if (changed.needsAnUpdate && (this.needsToRecompute !== false || this.needsToRevalidate !== false)) {
            changed.needsAnUpdate();
        }
    }
    deregisterValueChangedListener(changed) {
        const index = this.valueChangeListeners.findIndex(cb => cb === changed);
        if (index >= 0) {
            this.valueChangeListeners.splice(index, 1);
        }
    }
    tellValueChangeListeners(fcn) {
        this.valueChangeListeners.forEach(fcn);
    }
    setUpdateStrategy(strategy) {
        this.automaticallyUpdate = strategy === 'Automatic';
    }
    addUpdateTrigger(trigger, setUpdateStrategyManually = true) {
        if (setUpdateStrategyManually) {
            this.setUpdateStrategy("Manually");
        }
        trigger.registerTriggerListener(this.getTriggerListener());
    }
    removeUpdateTrigger(trigger) {
        if (this.triggerListener) {
            trigger.deregisterTriggerListener(this.triggerListener);
        }
    }
}
exports.AbstractPropertyImpl = AbstractPropertyImpl;
