import { alwaysAssertThat } from "../../util/assertions/assertions";
import { A } from "../../attributes/attribute-id";
export var TextInterpreter;
(function (TextInterpreter) {
    TextInterpreter[TextInterpreter["Plain"] = 0] = "Plain";
    TextInterpreter[TextInterpreter["Markdown"] = 1] = "Markdown";
    TextInterpreter[TextInterpreter["Html"] = 2] = "Html";
})(TextInterpreter || (TextInterpreter = {}));
export class PropertyScalarRuleBinding {
    constructor(property, notEmptyIfRequiredValidator, addDependencies) {
        this.notEmptyIfRequiredValidator = notEmptyIfRequiredValidator;
        this.addDependencies = addDependencies;
        this.property = property;
    }
    addScalarValidator(validator) {
        this.property.addScalarValidator(validator);
        return this;
    }
    addAsyncScalarValidator(validator) {
        const propList = [this.property];
        this.property.addValidator({
            getValidatedProperties: () => propList,
            validate: (prop) => validator(prop)
        });
        return this;
    }
    defineInitialValue(value) {
        alwaysAssertThat(!this.property.isReadOnly(), () => `${this.property.id}: Can not define initial value on read only property`);
        this.property.defineInitialValue(value);
        this.property.setToInitialValue();
        return this;
    }
    set(attribudeId, value) {
        this.defineAttributeFunction(attribudeId, [], () => value);
        return this;
    }
    define(attribudeId, ...dependencies) {
        return (value) => {
            this.defineAttributeFunction(attribudeId, dependencies, () => value(this.property, ...dependencies));
            return this;
        };
    }
    defineAttributeFunction(attribudeId, dependencies, fcn) {
        alwaysAssertThat(!A[attribudeId.name], () => {
            const mapping = {
                Required: 'defineRequiredIfVisible',
                Visible: 'defineVisibility',
                Label: 'defineLabel',
                InfoText: 'defineInfoText'
            };
            return `Please use specialised function ${mapping[attribudeId.name]} to define ${attribudeId.name}`;
        });
        this.addDependencies(dependencies, this.property, { [attribudeId.name]: true });
        this.property.defineAttribute({
            id: attribudeId,
            dependencies: dependencies,
            getValue: fcn
        });
    }
    setVisibility(isVisible) {
        if (isVisible === false) {
            this.defineVisibilityFunction([], () => isVisible);
        }
        else {
            this.property.defineVisibility(undefined);
        }
        return this;
    }
    defineVisibility(...dependencies) {
        return (isVisible) => {
            this.defineVisibilityFunction(dependencies, () => isVisible(this.property, ...dependencies));
            return this;
        };
    }
    defineVisibilityFunction(dependencies, fcn) {
        this.addDependencies(dependencies, this.property, { visible: true });
        this.property.defineVisibility({
            id: A.Visible,
            dependencies: dependencies,
            getValue: fcn
        });
    }
    setRequiredIfVisible(required) {
        if (required) {
            this.defineRequiredIfVisibleFunction([], () => required);
        }
        else {
            this.property.defineRequiredIfVisible(undefined);
        }
        return this;
    }
    defineRequiredIfVisible(...dependencies) {
        return (required) => {
            this.defineRequiredIfVisibleFunction(dependencies, () => required(this.property, ...dependencies));
            return this;
        };
    }
    defineRequiredIfVisibleFunction(dependencies, fcn) {
        this.addDependencies(dependencies, this.property, { required: true });
        this.property.defineRequiredIfVisible({
            id: A.Required,
            dependencies: dependencies,
            getValue: fcn
        });
        this.property.addScalarValidator(this.notEmptyIfRequiredValidator);
    }
    defineLabelAndPlaceholder(text) {
        this.defineLabel(text);
        return this.definePlaceholder(text);
    }
    defineLabel(text, textInterpreter) {
        this.property.defineLabel(this.interpreteText(text, textInterpreter));
        return this;
    }
    definePlaceholder(placeholder) {
        this.property.definePlaceholder(placeholder);
        return this;
    }
    defineInfoText(text, textInterpreter) {
        this.property.defineInfoText(this.interpreteText(text, textInterpreter));
        return this;
    }
    interpreteText(text, textInterpreter) {
        return text;
    }
    onUpdated(updated) {
        this.property.registerValueChangedListener({
            updated: () => updated(this.property)
        });
        return this;
    }
    setToInitialValueOnOtherPropertyChanged(...otherProperties) {
        alwaysAssertThat(!this.property.isReadOnly(), () => `${this.property.id}: Can only set properties to initial value, that are not read only`);
        return this;
    }
}
