export class StringValueConverter {
    fromDisplayValue(value) {
        return value;
    }
    asDisplayValue(value) {
        return value !== null && value !== void 0 ? value : '';
    }
    getNullFallbackValue() {
        return '';
    }
}
export class UpperCaseStringValueConverter {
    fromDisplayValue(value) {
        var _a;
        return (_a = value === null || value === void 0 ? void 0 : value.toUpperCase()) !== null && _a !== void 0 ? _a : null;
    }
    asDisplayValue(value) {
        return value !== null && value !== void 0 ? value : '';
    }
    getNullFallbackValue() {
        return '';
    }
}
export class NumberValueConverter {
    fromDisplayValue(value) {
        if (!value) {
            return null;
        }
        const converted = Number.parseFloat(value);
        return Number.isNaN(converted) ? null : converted;
    }
    asDisplayValue(value) {
        return value == null ? '' : value.toLocaleString();
    }
    getNullFallbackValue() {
        return 0;
    }
}
export class PrecisionNumberValueConverter {
    constructor(precision) {
        this.precision = precision;
    }
    fromDisplayValue(value) {
        if (!value) {
            return null;
        }
        const converted = Number.parseFloat(value);
        return Number.isNaN(converted) ? null : converted;
    }
    asDisplayValue(value) {
        return value == null ? '' : value.toFixed(this.precision);
    }
    getNullFallbackValue() {
        return 0;
    }
}
export class IntegerValueConverter {
    fromDisplayValue(value) {
        if (!value) {
            return null;
        }
        const converted = Number.parseInt(value);
        return Number.isNaN(converted) ? null : converted;
    }
    asDisplayValue(value) {
        return value == null ? '' : value.toFixed(0);
    }
    getNullFallbackValue() {
        return 0;
    }
}
export class BooleanConverter {
    constructor(trueString, falseString) {
        this.trueString = trueString;
        this.falseString = falseString;
        this.loweredTrueString = this.trueString.toLowerCase();
    }
    fromDisplayValue(value) {
        return (value === null || value === void 0 ? void 0 : value.toLowerCase()) === this.loweredTrueString;
    }
    asDisplayValue(value) {
        return value ? this.trueString : this.falseString;
    }
    getNullFallbackValue() {
        return false;
    }
}
function pad0(no) {
    return no < 10 ? `0${no}` : no;
}
function toDate(year, month, day) {
    if (!Number.isNaN(day) && !Number.isNaN(month) && !Number.isNaN(year)) {
        month--;
        const date = new Date(year, month, day);
        if (date.getFullYear() === year && date.getMonth() === month && date.getDate() === day) {
            return date;
        }
    }
    return null;
}
const currentYear = (new Date()).getFullYear();
const currentCentury = Math.floor(currentYear / 100) * 100;
const currentMonth = (new Date()).getMonth() + 1;
const digitsOnly = /^\d+$/;
export class IsoDateConverter {
    fromDisplayValue(value) {
        if (value) {
            let day = NaN;
            let month = NaN;
            let year = NaN;
            const splitted = value.split('-');
            if (splitted.length > 3 || splitted.some(split => !digitsOnly.test(split))) {
                return null;
            }
            if (splitted.length === 1) {
                value = splitted[0];
                if (value.length < 1 || value.length > 8 || (value.length !== 1 && value.length % 2 !== 0)) {
                    return null;
                }
                if (value.length === 8) {
                    year = Number.parseInt(value.substr(0, 4));
                    month = Number.parseInt(value.substr(4, 2));
                    day = Number.parseInt(value.substr(6, 2));
                }
                else if (value.length === 6) {
                    year = Number.parseInt(value.substr(0, 2)) + currentCentury;
                    month = Number.parseInt(value.substr(2, 2));
                    day = Number.parseInt(value.substr(4, 2));
                }
                else if (value.length === 4) {
                    year = currentYear;
                    month = Number.parseInt(value.substr(0, 2));
                    day = Number.parseInt(value.substr(2, 2));
                }
                else {
                    year = currentYear;
                    month = currentMonth;
                    day = Number.parseInt(value);
                }
            }
            else if (splitted.length === 2) {
                year = currentYear;
                if (splitted[0].length <= 2 && splitted[1].length <= 2) {
                    month = Number.parseInt(splitted[0]);
                    day = Number.parseInt(splitted[1]);
                }
            }
            else {
                if (splitted[0].length === 2) {
                    year = Number.parseInt(splitted[0]) + currentCentury;
                }
                else if (splitted[0].length === 4) {
                    year = Number.parseInt(splitted[0]);
                }
                if (splitted[1].length <= 2 && splitted[2].length <= 2) {
                    month = Number.parseInt(splitted[1]);
                    day = Number.parseInt(splitted[2]);
                }
            }
            return toDate(year, month, day);
        }
        return null;
    }
    asDisplayValue(value) {
        if (value) {
            const month = value.getMonth() + 1;
            return `${value.getFullYear()}-${pad0(month)}-${pad0(value.getDate())}`;
        }
        else {
            return '';
        }
    }
    getNullFallbackValue() {
        return new Date();
    }
}
export class GermanDateConverter {
    fromDisplayValue(value) {
        if (value) {
            let day = NaN;
            let month = NaN;
            let year = NaN;
            const splitted = value.split('.');
            if (splitted[splitted.length - 1] === '') {
                splitted.pop();
            }
            if (splitted.length > 3 || splitted.some(split => !digitsOnly.test(split))) {
                return null;
            }
            if (splitted.length === 1) {
                value = splitted[0];
                if (value.length < 1 || value.length > 8 || (value.length !== 1 && value.length % 2 !== 0)) {
                    return null;
                }
                day = Number.parseInt(value.substr(0, 2));
                if (value.length <= 2) {
                    month = currentMonth;
                    year = currentYear;
                }
                else {
                    month = Number.parseInt(value.substr(2, 2));
                }
                if (value.length === 4) {
                    year = currentYear;
                }
                else if (value.length === 6) {
                    year = Number.parseInt(value.substr(4, 2)) + currentCentury;
                }
                else if (value.length === 8) {
                    year = Number.parseInt(value.substr(4, 4));
                }
            }
            else {
                if (splitted[0].length <= 2 && splitted[1].length <= 2) {
                    day = Number.parseInt(splitted[0]);
                    month = Number.parseInt(splitted[1]);
                }
                if (splitted.length === 2) {
                    year = currentYear;
                }
                else if (splitted[2].length === 4) {
                    year = Number.parseInt(splitted[2]);
                }
                else if (splitted[2].length === 2) {
                    year = Number.parseInt(splitted[2]) + currentCentury;
                }
            }
            return toDate(year, month, day);
        }
        return null;
    }
    asDisplayValue(value) {
        if (value) {
            const month = value.getMonth() + 1;
            return `${pad0(value.getDate())}.${pad0(month)}.${value.getFullYear()}`;
        }
        else {
            return '';
        }
    }
    getNullFallbackValue() {
        return new Date();
    }
}
export const C = {
    string: {
        identity: new StringValueConverter(),
        upperCase: new UpperCaseStringValueConverter(),
    },
    number: {
        default: new NumberValueConverter(),
        withPrecision: (precision) => new PrecisionNumberValueConverter(precision),
        integer: new IntegerValueConverter(),
    },
    boolean: {
        default: new BooleanConverter('true', 'false'),
        jaNein: new BooleanConverter('Ja', 'Nein'),
    },
    date: {
        iso: new IsoDateConverter(),
        german: new GermanDateConverter(),
    }
};
